#include <stdio.h>
#include <starlet.h>
#include <unixio.h>
#include <lib$routines.h>
#include <stat.h>
#include <string.h>
#include <stdlib.h>
#include <rmsdef.h>
#include <fab.h>
#include <nam.h>

static int directory_file_name (char *, char *);

/*
 * Not defined on VMS unless you include <socket.h>
*/

struct timeval {
    long	tv_sec;
    long	tv_usec;
};

#define	MAX_EVENT_FLAGS		32

/*
 * Try and avoid using this event flag, we use it to set a timer.
 *
 * The X windows event flag has to be in the same cluster as this flag.
 */
static int      TIMER_EVENT_FLAG = 0;

static void CvtTimeval2Quad(struct timeval *, long *);

/*
 * Bitmap for event flags
 */
static unsigned long	eventFlagMap = 0;

static int	eventFlagInit = 1;

reserveEventFlag(int flag)
{
    /*
     * Mark an event flag in use
     */
    unsigned long mask;

    if (flag >= MAX_EVENT_FLAGS)
	panic("Can't handle event flag %d", flag);
    mask = (1 << flag);
    eventFlagMap |= mask;
}

getEventFlag()
{
    /*
     * Return an event flag
     *
     * 0 is never used.
     * 1 is the timer event flag
     */
    int i, j;

    unsigned long mask;

    if (eventFlagInit) {
	for (i=1; i<MAX_EVENT_FLAGS; i++) {
            lib$free_ef(&i);
	}
	eventFlagInit = 0;
    }

    for (i=1; i<MAX_EVENT_FLAGS; i++) {
	mask = (1 << i);
	if (!(eventFlagMap & mask)) {
	    if (lib$reserve_ef(&i) & 1) {
	    	eventFlagMap |= mask;
	    	return i;
	    }
	}
    }
    fprintf(stderr, "GetEventFlag: No free event flags.\n");
    return 0;
}

void
freeEventFlag(int flag)
{
    /*
     * Return an event to the pool
     */
    unsigned long mask;

    mask = (1 << flag);
    if (eventFlagMap & mask) {
	lib$free_ef(&flag);
	eventFlagMap &= ~mask;
    }
    else
	fprintf(stderr, "freeEventFlag: Flag %d not reserved.\n", flag);
}

vmsSelect(int nDesc, int *rdsc, int *wdsc, int *edsc, struct timeval *timeOut)
{
    /*
     * Emulate a select call
     *
     * 0 is never used.
     * 1 is our timer event flag
     * 2-31 the rest
     */
    int nfds = 0;
    int status;
    int poll;
    int timerSet = 0;
    long quadTime[2];
    unsigned long mask, readMask, waitMask;
    unsigned long baseMask;

    if (TIMER_EVENT_FLAG == 0) {
	TIMER_EVENT_FLAG = getEventFlag();
    }
    poll = timeOut && timeOut->tv_sec == 0 && timeOut->tv_usec == 0;
	
    sys$setast(0);			/* Block interrupts */
    
    if (rdsc)
	readMask = *rdsc;
    else
	readMask = 0;			/* Must be a wait call */

    waitMask = readMask;
    if (!poll && timeOut) {		/* Do we need to start a timer? */
	CvtTimeval2Quad(timeOut, quadTime);
	sys$cantim(1, 0);
	if (!((status = sys$setimr(TIMER_EVENT_FLAG, quadTime, 0, 1)) & 1)) {
		poll = 0;
	}
	waitMask |= (1 << TIMER_EVENT_FLAG);
	timerSet = 1;
    } else {
	timerSet = 0;
    }
    
    sys$setast(1);			/* Re-enable interrupts */
    
    if (!poll) {
        sys$readef(TIMER_EVENT_FLAG, &mask);
	sys$wflor(TIMER_EVENT_FLAG, waitMask);
    }
    sys$readef(TIMER_EVENT_FLAG, &mask);

    /*
     * Count number of descriptors that are ready
     */
    mask &= readMask;			/* Only count those we care about */
    if (rdsc) {
	*rdsc = mask;
	for (nfds = 0; mask; mask >>= 1) {
	    if (mask & 1)
		nfds++;
	}
    }
    else
    	nfds = 0;

    /*
    	Don't do these on VMS
    */
    if (wdsc)
	*wdsc = 0;
    if (edsc)
	*edsc = 0;
    return(nfds);
}

static void
CvtTimeval2Quad(struct timeval *tv, long *quad)
{
  /*
   * Convert a Unix time to VMS quad time
   */
  int multiplier = - 10*1000*1000;
  int addend = - tv->tv_usec * 10;

  lib$emul (&multiplier, &tv->tv_sec, &addend, quad);
}

VMSaccess(char *name, int what)
{
    char newName[NAM$C_MAXRSS];

    if (directory_file_name(name, newName))
    	return access(newName, what);
    else
    	return -1;
}

VMSstat(char *name, stat_t *statBuff)
{
    char newName[NAM$C_MAXRSS];

    if (strpbrk(name, "?*"))
    	return -1;
    if (directory_file_name(name, newName))
	return stat(newName, statBuff);
    else
	return -1;
}

/*
 * Convert from directory name to filename.
 * On VMS:
 *       xyzzy:[mukesh.emacs] => xyzzy:[mukesh]emacs.dir.1
 *       xyzzy:[mukesh] => xyzzy:[000000]mukesh.dir.1
 * On UNIX, it's simple: just make sure there is a terminating /

 * Value is nonzero if the string output is different from the input.
 *
 * From GNU Emacs
 */

static
directory_file_name (char *src, char *dst)
{
    char		esa[NAM$C_MAXRSS];
    char 		*ptr, *rptr;
    char		bracket;
    long		slen;
    long		rlen;
    struct		FAB fab = cc$rms_fab;
    struct		NAM nam = cc$rms_nam;

    slen = strlen (src) - 1;
    if ((src[slen] == ']' || src[slen] == ':' || src[slen] == '>')) {
	/* VMS style - convert [x.y.z] to [x.y]z, [x] to [000000]x */
	fab.fab$l_fna = src;
	fab.fab$b_fns = slen + 1;
	fab.fab$l_nam = &nam;
	fab.fab$l_fop = FAB$M_NAM;

	nam.nam$l_esa = esa;
	nam.nam$b_ess = sizeof esa;
	nam.nam$b_nop |= NAM$M_SYNCHK;

	/* We call SYS$PARSE to handle such things as [--] for us. */
	if (sys$parse(&fab, 0, 0) == RMS$_NORMAL) {
	    slen = nam.nam$b_esl - 1;
	    if (esa[slen] == ';' && esa[slen - 1] == '.')
		slen -= 2;
	    esa[slen + 1] = '\0';
	    src = esa;
	}
	if (src[slen] != ']' && src[slen] != '>') {
	    /* what about when we have logical_name:???? */
	    if (src[slen] == ':') {	/* Xlate logical name and see what we get */
		ptr = strcpy (dst, src); /* upper case for getenv */
		while (*ptr) {
		    if ('a' <= *ptr && *ptr <= 'z')
			*ptr -= 040;
		    ptr++;
		}
		dst[slen] = 0;	/* remove colon */
		if (!(src = getenv (dst)))
		    return 0;
		/* should we jump to the beginning of this procedure?
		 Good points: allows us to use logical names that xlate
		 to Unix names,
		 Bad points: can be a problem if we just translated to a device
		 name...
		 For now, I'll punt and always expect VMS names, and hope for
		 the best! */
		slen = strlen (src) - 1;
		if (src[slen] != ']' && src[slen] != '>') { /* no recursion here! */
		    strcpy (dst, src);
		    return 0;
		}
	    }
	    else {		/* not a directory spec */
		strcpy (dst, src);
		return 0;
	    }
	}
	bracket = src[slen];
	if (!(ptr = strchr (src, bracket - 2))) { /* no opening bracket */
	    strcpy (dst, src);
	    return 0;
	}
	if (!(rptr = strrchr (src, '.')))
	    rptr = ptr;
	slen = rptr - src;
	strncpy (dst, src, slen);
	dst[slen] = '\0';
	if (*rptr == '.') {
	    dst[slen++] = bracket;
	    dst[slen] = '\0';
	}
	else {
	    /*
	      If we have the top-level of a rooted directory
	      (i.e. xx:[000000]), then translate the device and recurse.
	    */
	    if (dst[slen - 1] == ':'
		&& dst[slen - 2] != ':'	/* skip decnet nodes */
		&& strcmp(src + slen, "[000000]") == 0) {
		dst[slen - 1] = '\0';
		if ((ptr = getenv (dst))
		  && (rlen = strlen (ptr) - 1) > 0
		  && (ptr[rlen] == ']' || ptr[rlen] == '>')
		  && ptr[rlen - 1] == '.') {
		    ptr[rlen - 1] = ']';
		    ptr[rlen] = '\0';
		    return directory_file_name (ptr, dst);
		}
		else
		    dst[slen - 1] = ':';
	    }
	    strcat (dst, "[000000]");
	    slen += 8;
	}
	rptr++;
	rlen = strlen (rptr) - 1;
	strncat (dst, rptr, rlen);
	dst[slen + rlen] = '\0';
	strcat (dst, ".DIR.1");
	return 1;
    }

    strcpy (dst, src);
    return 1;
}
