#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stat.h>
#include <unixlib.h>

#define	eq(x,y)	(strcmp(x,y) == 0)

static char *copyComponent(char *, char *);
static char *addString(char *, char *);
static int isDirectory(char *);
static int trname(char *, char *);
static int foldUp(char *, char *);

Tcl_UnixToVms(char *unixFile, char *vmsFile)
{
    /*
     * Convert a Unix filename to VMS
     */
    char *u;
    char *v;
    char *s;
    char unxFile[256];
    char component[256];
    char translated[256];
    char *saveV, saveC;
    int	 l;

    u = unxFile;
    v = vmsFile;
    *v = 0;
    strcpy(u, unixFile);
    l = strlen(u);
    if (l > 3) {
    	if (unixFile[l-1] == '.' && unixFile[l-2] == '.')
    	    strcat(u, "/");
	else if (unixFile[l-1] == '.' && unixFile[l-2] == '/')
    	    strcat(u, "/");
    }
    if (*u == '~') {
	u = copyComponent(u+1, component);
	if (!*component)
	    v = addString(getenv("HOME"), v);
	else
	    return(0);
    }
    else if (*u == '/') {
	component[0] = 0;
	if ((u = copyComponent(u+1, component)) || *component) {
	    if (trname(component, translated)) {
	    	l = strlen(translated);
	    	v = addString(translated, v);
	    	if (l > 2) {
		    if (translated[l-1] == ']' && translated[l-2] == '.') {
		    	*(v-2) = ']';
		    	*(v-1) = 0;
		    	v--;
		    }
		    else if (translated[l-1] == ':')
		    	;
		}
	    }
	    else {
		v = addString(component, v);
		v = addString(":", v);
	    }
	}
	else					/* Just `/' */
	    return(0);				/*  can not handle that */
    }
    else if (*u == '.') {
	if (*(u+1) == '/') {
	    u += 2;				/* Ignore ./ */
	    if (*u == 0)			/* Just ./ */
		strcpy(v, "[]");		/* All done */
	}
	else if (*(u+1) == '.') {		/* ../name/... */
	    if (*(u+2) == '/' || *(u+2) == 0) {
		v = addString("[-]", v);
		u += *(u+2) ? 3 : 2;
	    }
	    else {				/* It is ..name */
		u = copyComponent(u, component);
		*v++ = '[';
		v = addString(component, v);
		v = addString("]", v);
	    }
	}
	else if (*(u+1) == 0) {
	    strcpy(v, "[]");
	    u++;
	}
	else {
	    u = copyComponent(u, component);
	    if (u) {				/* .name/... */
		*v++ = '[';
		v = addString(component, v);
		v = addString("]", v);
	    }
	    else				/* only .name */
		v = addString(component, v);
	}
    }
    else {
	/*
	 * See if we have a VMS/Unix hybrid name
	 */
	s = copyComponent(u, component);
	l = strlen(component);
	if (component[l-1] == ':' || component[l-1] == ']') {
	    v = addString(component, v);
	    u = s;
	}
    }

    /*
     * The rest of the string is of the form: name1/name2/..../namen
     */
    while (u != NULL && *u != 0) {
	if (u = copyComponent(u, component)) {
	    if (eq(component, ".."))
		strcpy(component, "-");
	    else if (eq(component, "."))
		continue;
	    if (v == vmsFile) {
		*v++ = '[';
		*v++ = '.';
	    }
	    else if (*(v-1) == ':')
		*v++ = '[';
	    else
		*(v-1) = '.';
	    v = addString(component, v);
	    v = addString("]", v);
	}
	else {					/* last part of unix pathname */
	    saveV = v;
	    if (v != vmsFile && *(v-1) == ':') {
		*(v-1) = 0;
		if (trname(vmsFile, translated) &&
		    strchr(translated, ']') &&
		    strstr(translated, ".]"))
		    *(v-1) = ':';
		else {
		    *(v-1) = ':';
		    v = addString("[000000]", v);
		}
	    }
	    v = addString(component, v);
	    if (strchr(component, '.'))
		break;				/* Explicit extension */
	    v = addString(".dir", v);		/* Check for directory */
	    if (isDirectory(vmsFile)) {
		if (saveV == vmsFile) {
		    v = saveV;
		    *v++ = '[';
		    *v++ = '.';
		}
		else {
		    v = saveV-1;
		    if (*v == ':') {
			v++;
			*v++ = '[';
		    }
		    else
			*v++ = '.';
		}
		v = addString(component, v);
		v = addString("]", v);
	    }
	    else
		v = addString(component, saveV);
	    break;
	}
    }
    return(1);
}
static char *
copyComponent(char *unixFile, char *component)
{
    /*
     * Return the next component from a Unix file spec
     *
     * Remove ".dir" from the string if present.
     */
    char *sc = component;
    char c;
    char *t, tcomp[256];

    if (unixFile) {
	while ((c = *unixFile++) && c != '/')
	    *component++ = c;
	if (c != '/')
	    unixFile--;
	*component = 0;
	foldUp(sc, tcomp);
	if (t = strstr(tcomp, ".DIR"))
	    sc[t-tcomp] = 0;
	return(c ? unixFile : NULL);
    }
    else
	return(NULL);
}

static char *
addString(char *from, char *to)
{
    while (*to++ = *from++)
	;
    return(--to);
}

static
isDirectory(char *file)
{
    /*
     * Return TRUE if `file' points to a directory
     */
    struct stat statBuff;

    return((stat(file, &statBuff) == 0) && ((statBuff.st_mode & S_IFDIR) != 0));
}

static
trname(char *in, char *out)
{
    char *s;
    char *getenv();
    char tin[256];

    foldUp(in, tin);
    if (s = getenv(tin)) {
	strcpy(out, s);
	return 1;
    }
    else
	return 0;
}

static
foldUp(char *in, char *out)
{
    char	c;

    while (c = *in++) {
        if (islower(c))
            c = _toupper(c);
        *out++ = c;
    }
    *out = 0;
}

void
Tcl_VmsToUnix(char *vmsFile, char *unixFile)
{
    /*
     *	Convert a filename in VMS format to Unix format
     */
    char *t, *u;

    if (strpbrk(vmsFile, "]>:")) {
	u = decc$translate_vms(vmsFile);

	/*
	 *  Work around DEC C bug
	 */
	if (vmsFile[0] == '[' && vmsFile[1] == '.' && *u == '/')
	    u++;
	if (*u == '.' && *(u+1) == '/')
	    u += 2;
	t = u;
    }
    else
	t = vmsFile;

    strcpy(unixFile, t);
}

